import numpy as np
from tensorflow.keras import layers
from tensorflow.keras.models import Model

from ..layers.InstanceNorm import InstanceNormalization


def vox_gan(
    n_classes,
    input_shape,
    g_filters=64,
    g_kernel_size=4,
    g_norm="batch",
    d_filters=64,
    d_kernel_size=4,
    d_norm="batch",
):
    """Instantiate Vox2VoxGAN.

    Adapted from https://arxiv.org/abs/2003.13653
    Code: https://github.com/mdciri/Vox2Vox

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    g_kernal_size: int, size of the kernal for generator. Default kernal size
        is set to be 4.
    g_filters: int, number of filters for generator. default is set 64.
    g_norm: str, to set batch or instance norm.
    d_kernal_size: int, size of the kernal for discriminator. Default kernal size
        is set to be 4.
    d_filters: int, number of filters for discriminator. default is set 64.
    d_norm: str, to set batch or instance norm.

    Returns
    ----------
    Model object.

    """
    generator = Vox_generator(
        n_classes,
        input_shape,
        n_filters=g_filters,
        kernel_size=g_kernel_size,
        norm=g_norm,
    )

    discriminator = Vox_discriminator(
        input_shape, n_filters=d_filters, kernel_size=d_kernel_size, norm=d_norm
    )
    return generator, discriminator


def Vox_generator(n_classes, input_shape, n_filters=64, kernel_size=4, norm="batch"):
    """Instantiate Generator.

    Adapted from https://arxiv.org/abs/2003.13653

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 4.
    n_filters: int, number of filters. default is set 64.
    norm: str, to set batch or instance norm.

    Returns
    ----------
    Model object.

    """

    def encoder_step(inputs, filters, kernel_size=3, norm="instance"):
        x = layers.Conv3D(
            filters,
            kernel_size=kernel_size,
            strides=2,
            kernel_initializer="he_normal",
            padding="same",
        )(inputs)
        if norm == "instance":
            x = InstanceNormalization()(x)
        if norm == "batch":
            x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)
        x = layers.Dropout(0.2)(x)

        return x

    def bottleneck(inputs, filters, kernel_size, norm="instance"):
        x = layers.Conv3D(
            filters,
            kernel_size=kernel_size,
            strides=2,
            kernel_initializer="he_normal",
            padding="same",
        )(inputs)
        if norm == "instance":
            x = InstanceNormalization()(x)
        if norm == "batch":
            x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)
        for i in range(4):
            y = layers.Conv3D(
                filters,
                kernel_size=kernel_size,
                strides=1,
                kernel_initializer="he_normal",
                padding="same",
            )(x)
            if norm == "instance":
                x = InstanceNormalization()(y)
            if norm == "batch":
                x = layers.BatchNormalization()(y)
            x = layers.LeakyReLU()(x)
            x = layers.Concatenate()([x, y])

        return x

    def decoder_step(
        inputs, layer_to_concatenate, filters, kernel_size, norm="instance"
    ):
        x = layers.Conv3DTranspose(
            filters,
            kernel_size,
            strides=2,
            padding="same",
            kernel_initializer="he_normal",
        )(inputs)
        if norm == "instance":
            x = InstanceNormalization()(x)
        if norm == "batch":
            x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)
        x = layers.Concatenate()([x, layer_to_concatenate])
        x = layers.Dropout(0.2)(x)
        return x

    layers_to_concatenate = []
    inputs = layers.Input(input_shape, name="input_image")
    Nfilter_start = n_filters
    depth = 4
    x = inputs

    # encoder
    for d in range(depth - 1):
        if d == 0:
            x = encoder_step(
                x, Nfilter_start * np.power(2, d), kernel_size, norm="None"
            )
        else:
            x = encoder_step(x, Nfilter_start * np.power(2, d), kernel_size, norm=norm)
        layers_to_concatenate.append(x)

    # bottlenek
    x = bottleneck(x, Nfilter_start * np.power(2, depth - 1), kernel_size, norm=norm)

    # decoder
    for d in range(depth - 2, -1, -1):
        x = decoder_step(
            x,
            layers_to_concatenate.pop(),
            Nfilter_start * np.power(2, d),
            kernel_size,
            norm=norm,
        )

    # classifier
    last = layers.Conv3DTranspose(
        n_classes,
        kernel_size=kernel_size,
        strides=2,
        padding="same",
        kernel_initializer="he_normal",
        activation="softmax",
        name="output_generator",
    )(x)

    return Model(inputs=inputs, outputs=last, name="Generator")


def Vox_discriminator(input_shape, n_filters=64, kernel_size=4, norm="batch"):
    """Instantiate Discriminator.

    Adapted from https://arxiv.org/abs/2003.13653

    Parameters
    ----------
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    n_filters: int, number of filters. default is set 64.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 4.
    norm: str, to set batch or instance norm.

    Returns
    ----------
    Model object.

    """

    inputs = layers.Input(input_shape, name="input_image")
    targets = layers.Input(input_shape, name="target_image")
    Nfilter_start = n_filters
    depth = 3

    def encoder_step(inputs, n_filters, kernel_size, norm="instance"):
        x = layers.Conv3D(
            n_filters,
            kernel_size,
            strides=2,
            kernel_initializer="he_normal",
            padding="same",
        )(inputs)
        if norm == "instance":
            x = InstanceNormalization()(x)
        if norm == "batch":
            x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)
        x = layers.Dropout(0.2)(x)

        return x

    x = layers.Concatenate()([inputs, targets])

    for d in range(depth):
        if d == 0:
            x = encoder_step(
                x, Nfilter_start * np.power(2, d), kernel_size, norm="None"
            )
        else:
            x = encoder_step(x, Nfilter_start * np.power(2, d), kernel_size, norm=norm)

    x = layers.ZeroPadding3D()(x)
    x = layers.Conv3D(
        Nfilter_start * (2**depth),
        kernel_size,
        strides=1,
        padding="valid",
        kernel_initializer="he_normal",
    )(x)
    if norm == "instance":
        x = InstanceNormalization()(x)
    if norm == "batch":
        x = layers.BatchNormalization()(x)
    x = layers.LeakyReLU()(x)

    x = layers.ZeroPadding3D()(x)
    last = layers.Conv3D(
        1,
        kernel_size,
        strides=1,
        padding="valid",
        kernel_initializer="he_normal",
        name="output_discriminator",
    )(x)

    return Model(inputs=[targets, inputs], outputs=last, name="Discriminator")


def Vox_ensembler(n_classes, input_shape, kernel_size=3, **kwargs):
    """Instantiate Ensembler.

    Adapted from https://arxiv.org/abs/2003.13653

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 3.

    Returns
    ----------
    Model object.
    """
    start = layers.Input(input_shape)
    fin = layers.Conv3D(
        n_classes,
        kernel_size=kernel_size,
        kernel_initializer="he_normal",
        padding="same",
        activation="softmax",
    )(start)

    return Model(inputs=start, outputs=fin, name="Ensembler")



"""Methods for creating `tf.data.Dataset` objects."""

import math
import os
from pathlib import Path

import fsspec
import nibabel as nb
import numpy as np
import tensorflow as tf

from .io import _is_gzipped, verify_features_labels
from .tfrecord import _labels_all_scalar, parse_example_fn, write
from .volume import binarize, replace, standardize, to_blocks

AUTOTUNE = tf.data.experimental.AUTOTUNE


def tfrecord_dataset(
    file_pattern,
    volume_shape,
    shuffle,
    scalar_label,
    compressed=True,
    num_parallel_calls=AUTOTUNE,
):
    """Return `tf.data.Dataset` from TFRecord files."""
    dataset = tf.data.Dataset.list_files(file_pattern, shuffle=shuffle)
    # Read each of these files as a TFRecordDataset.
    # Assume all files have same compression type as the first file.
    compression_type = "GZIP" if compressed else None
    cycle_length = 1 if num_parallel_calls is None else num_parallel_calls
    parse_fn = parse_example_fn(volume_shape=volume_shape, scalar_label=scalar_label)

    if not shuffle:
        # Determine examples_per_shard from the first TFRecord shard
        # Then set block_length to equal the number of examples per shard
        # so that the interleave method does not inadvertently shuffle data.
        first_shard = (
            dataset.take(1)
            .flat_map(
                lambda x: tf.data.TFRecordDataset(x, compression_type=compression_type)
            )
            .map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
        )
        block_length = len([0 for _ in first_shard])
    else:
        # If the dataset is being shuffled, then we don't care if interleave
        # further shuffles that data even further
        block_length = None

    dataset = dataset.interleave(
        map_func=lambda x: tf.data.TFRecordDataset(
            x, compression_type=compression_type
        ),
        cycle_length=cycle_length,
        block_length=block_length,
        num_parallel_calls=num_parallel_calls,
    )
    dataset = dataset.map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
    return dataset


def get_dataset(
    file_pattern,
    n_classes,
    batch_size,
    volume_shape,
    scalar_label=False,
    block_shape=None,
    n_epochs=None,
    mapping=None,
    augment=None,
    normalizer=standardize,
    shuffle_buffer_size=None,
    num_parallel_calls=AUTOTUNE,
):
    """Return `tf.data.Dataset` that preprocesses data for training or prediction.

    Labels are preprocessed for binary or multiclass segmentation according to
    `n_classes`.

    Parameters
    ----------
    file_pattern: str, expression that can be globbed to get TFRecords files
        for this dataset. For example 'data/training_*.tfrecords'.
    n_classes: int, number of classes to segment. Values of 1 and 2 indicate
        binary segmentation (foreground vs background), and values greater than
        2 indicate multiclass segmentation.
    batch_size: int, number of elements per batch.
    volume_shape: tuple of at least length 3, the shape of every volume in the TFRecords
        files. Every volume must have the same shape.
    scalar_label: boolean, if `True`, labels are scalars.
    block_shape: tuple of at least length 3, the shape of the non-overlapping sub-volumes
        to take from the full volumes. If None, do not separate the full volumes
        into sub-volumes. Separating into non-overlapping sub-volumes is useful
        (sometimes even necessary) to overcome memory limitations depending on
        the number of model parameters.
    n_epochs: int, number of epochs for the dataset to repeat. If None, the
        dataset will be repeated indefinitely.
    mapping: dict, mapping to replace label values. Values equal to a key in
        the mapping are replaced with the corresponding values in the mapping.
        Values not in `mapping.keys()` are replaced with zeros.
    augment: None, or list of different transforms in the executable sequence
            the corresponding arguments in tuple as e.g.:
            [(addGaussianNoise, {'noise_mean':0.1,'noise_std':0.5}), (...)]
    normalizer: callable, applies this normalization function when creating the
        dataset. to maintain compatibility with prior nobrainer release, this is
        set to standardize by default.
    shuffle_buffer_size: int, buffer of full volumes to shuffle. If this is not
        None, then the list of files found by 'file_pattern' is also shuffled
        at every iteration.
    num_parallel_calls: int, number of parallel calls to make for data loading
        and processing.

    Returns
    -------
    `tf.data.Dataset` of features and labels. If block_shape is not None, the
    shape of features is `(batch_size, *block_shape, 1)` and the shape of labels
    is `(batch_size, *block_shape, n_classes)`. If block_shape is None, then
    the shape of features is `(batch_size, *volume_shape, 1)` and the shape of
    labels is `(batch_size, *volume_shape, n_classes)`. If `scalar_label` is `True,
    the shape of labels is always `(batch_size,)`.
    """

    fs, _, _ = fsspec.get_fs_token_paths(file_pattern)
    files = fs.glob(file_pattern)
    if not files:
        raise ValueError("no files found for pattern '{}'".format(file_pattern))

    # Create dataset of all TFRecord files. After this point, the dataset will have
    # two value per iteration: (feature, label).
    shuffle = bool(shuffle_buffer_size)
    compressed = _is_gzipped(files[0], filesys=fs)
    dataset = tfrecord_dataset(
        file_pattern=file_pattern,
        volume_shape=volume_shape,
        shuffle=shuffle,
        scalar_label=scalar_label,
        compressed=compressed,
        num_parallel_calls=num_parallel_calls,
    )

    if normalizer is not None:
        # Standard-score the features.
        dataset = dataset.map(lambda x, y: (normalizer(x), y))

    # Augment examples if requested.
    if isinstance(augment, bool):
        raise ValueError("Augment no longer supports a boolean expression")

    if augment is not None:
        for transform, kwargs in augment:
            dataset = dataset.map(
                lambda x, y: tf.cond(
                    tf.random.uniform((1,)) > 0.5,
                    true_fn=lambda: transform(x, y, **kwargs),
                    false_fn=lambda: (x, y),
                ),
                num_parallel_calls=num_parallel_calls,
            )

    # Separate into blocks, if requested.
    if block_shape is not None:
        if not scalar_label:
            dataset = dataset.map(
                lambda x, y: (to_blocks(x, block_shape), to_blocks(y, block_shape)),
                num_parallel_calls=num_parallel_calls,
            )
            # This step is necessary because separating into blocks adds a dimension.
            dataset = dataset.unbatch()
        if scalar_label:

            def _f(x, y):
                x = to_blocks(x, block_shape)
                n_blocks = x.shape[0]
                y = tf.repeat(y, n_blocks)
                return (x, y)

            dataset = dataset.map(_f, num_parallel_calls=num_parallel_calls)
            # This step is necessary because separating into blocks adds a dimension.
            dataset = dataset.unbatch()
    else:
        if scalar_label:
            dataset = dataset.map(lambda x, y: (x, tf.squeeze(y)))

    # Binarize or replace labels according to mapping.
    if not scalar_label:
        if n_classes < 1:
            raise ValueError("n_classes must be > 0.")
        elif n_classes == 1:
            dataset = dataset.map(lambda x, y: (x, tf.expand_dims(binarize(y), -1)))
        elif n_classes == 2:
            dataset = dataset.map(lambda x, y: (x, tf.one_hot(binarize(y), n_classes)))
        elif n_classes > 2:
            if mapping is not None:
                dataset = dataset.map(lambda x, y: (x, replace(y, mapping=mapping)))
            dataset = dataset.map(lambda x, y: (x, tf.one_hot(y, n_classes)))

    # If volume_shape is only three dims, add grayscale channel to features.
    # Otherwise, assume that the channels are already in the features.
    if len(volume_shape) == 3:
        dataset = dataset.map(lambda x, y: (tf.expand_dims(x, -1), y))

    # Prefetch data to overlap data production with data consumption. The
    # TensorFlow documentation suggests prefetching `batch_size` elements.
    dataset = dataset.prefetch(buffer_size=batch_size)

    # Batch the dataset, so each iteration gives `batch_size` elements. We drop
    # the remainder so that when training on multiple GPUs, the batch will
    # always be evenly divisible by the number of GPUs. Otherwise, the last
    # batch might have fewer than `batch_size` elements and will cause errors.
    if batch_size is not None:
        dataset = dataset.batch(batch_size=batch_size, drop_remainder=True)

    # Optionally shuffle. We also optionally shuffle the list of files.
    # The TensorFlow recommend shuffling and then repeating.
    if shuffle_buffer_size:
        dataset = dataset.shuffle(buffer_size=shuffle_buffer_size)

    # Repeat the dataset for n_epochs. If n_epochs is None, then repeat
    # indefinitely. If n_epochs is 1, then the dataset will only be iterated
    # through once.
    dataset = dataset.repeat(n_epochs)

    return dataset


def get_steps_per_epoch(n_volumes, volume_shape, block_shape, batch_size):
    def get_n(a, k):
        return (a - k) / k + 1

    n_blocks = tuple(get_n(aa, kk) for aa, kk in zip(volume_shape, block_shape))

    for n in n_blocks:
        if not n.is_integer() or n < 1:
            raise ValueError(
                "cannot create non-overlapping blocks with the given parameters."
            )
    n_blocks_per_volume = np.prod(n_blocks).astype(int)

    steps = n_blocks_per_volume * n_volumes / batch_size
    steps = math.ceil(steps)
    return steps


def write_multi_resolution(
    paths,
    tfrecdir=Path(os.getcwd()) / "data",
    resolutions=None,
    shard_size=3,
    n_processes=1,
):
    """Create a multiresolution dataset.

    This returns a dictionary of information that is used by the generative model
    processing class.

    TODO: This function needs to be aligned with the Dataset class
    """
    resolutions = resolutions or [8, 16, 32, 64, 128, 256]
    tfrecdir = Path(tfrecdir)
    tfrecdir.mkdir(exist_ok=True, parents=True)
    template = tfrecdir / "data-train_shard-{shard:03d}.tfrec"

    write(
        features_labels=paths,
        filename_template=str(template),
        examples_per_shard=shard_size,  # change for larger dataset
        multi_resolution=True,
        resolutions=resolutions,
        processes=n_processes,
    )

    datasets = {}
    for resolution in resolutions:
        datasets[resolution] = dict(
            file_pattern=str(tfrecdir / f"*res-{resolution:03d}.tfrec"),
            batch_size=1,
            normalizer=None,
        )
    return datasets


class Dataset:
    """Datasets for training, and validation"""

    def __init__(
        self, n_classes, batch_size, block_shape, volume_shape=None, n_epochs: int = 1
    ):
        self.n_classes = n_classes
        self.volume_shape = volume_shape
        self.block_shape = block_shape
        self.batch_size = batch_size
        self.n_epochs = n_epochs

    def from_tfrecords(
        self,
        volume_shape,
        scalar_labels,
        n_volumes,
        template="data/data-train_shard-*.tfrec",
        augment=None,
        shuffle_buffer_size=None,
        num_parallel_calls=1,
    ):
        """Function to retrieve a saved tf record as a Dataset

        template: str, the path to which TFRecord files should be written.
        num_parallel_calls: int, number of processes to use for multiprocessing. If
            None, will use all available processes.
        """
        self.volume_shape = volume_shape

        # replace shard formatting code with * for globbing
        dataset = get_dataset(
            file_pattern=template,
            n_classes=self.n_classes,
            batch_size=self.batch_size,
            volume_shape=self.volume_shape,
            block_shape=self.block_shape,
            n_epochs=self.n_epochs,
            augment=augment,
            shuffle_buffer_size=shuffle_buffer_size,
            num_parallel_calls=num_parallel_calls,
        )
        # Add nobrainer specific attributes
        dataset.scalar_labels = scalar_labels
        dataset.n_volumes = n_volumes
        dataset.volume_shape = self.volume_shape
        return dataset

    def from_files(
        self,
        paths,
        eval_size=0.1,
        tfrecdir=Path(os.getcwd()) / "data",
        shard_size=3,
        augment=None,
        shuffle_buffer_size=None,
        num_parallel_calls=1,
        check_shape=True,
        check_labels_int=False,
        check_labels_gte_zero=False,
    ):
        """Create Nobrainer datasets from data

        template: str, the path to which TFRecord files should be written. A string
            formatting key `shard` should be included to indicate the unique TFRecord file
            when writing to multiple TFRecord files. For example,
            `data_shard-{shard:03d}.tfrec`.
        shard_size: int, number of pairs of `(feature, label)` per TFRecord file.
        check_shape: boolean, if true, validate that the shape of both volumes is
            equal to 'volume_shape'.
        check_labels_int: boolean, if true, validate that every labels volume is an
            integer type or can be safely converted to an integer type.
        check_labels_gte_zero: boolean, if true, validate that every labels volume
            has values greater than or equal to zero.
        num_parallel_calls: int, number of processes to use for multiprocessing. If
            None, will use all available processes.
        """
        # Test that the `filename_template` has a `shard` formatting key.
        template = str(Path(tfrecdir) / "data-{intent}")
        shard_ext = "shard-{shard:03d}.tfrec"

        Neval = np.ceil(len(paths) * eval_size).astype(int)
        Ntrain = len(paths) - Neval

        verify_result = verify_features_labels(
            paths,
            check_shape=check_shape,
            check_labels_int=check_labels_int,
            check_labels_gte_zero=check_labels_gte_zero,
        )
        if len(verify_result) == 0:
            Path(tfrecdir).mkdir(exist_ok=True, parents=True)
            if self.volume_shape is None:
                self.volume_shape = nb.load(paths[0][0]).shape
            write(
                features_labels=paths[:Ntrain],
                filename_template=template.format(intent=f"train_{shard_ext}"),
                examples_per_shard=shard_size,
                processes=num_parallel_calls,
            )
            if Neval > 0:
                write(
                    features_labels=paths[Ntrain:],
                    filename_template=template.format(intent=f"eval_{shard_ext}"),
                    examples_per_shard=shard_size,
                    processes=num_parallel_calls,
                )
            labels = (y for _, y in paths)
            scalar_labels = _labels_all_scalar(labels)
            # replace shard formatting code with * for globbing
            template_train = template.format(intent="train_*.tfrec")
            ds_train = self.from_tfrecords(
                self.volume_shape,
                scalar_labels,
                len(paths[:Ntrain]),
                template=template_train,
                augment=augment,
                shuffle_buffer_size=shuffle_buffer_size,
                num_parallel_calls=num_parallel_calls,
            )
            ds_eval = None
            if Neval > 0:
                template_eval = template.format(intent="eval_*.tfrec")
                ds_eval = self.from_tfrecords(
                    self.volume_shape,
                    scalar_labels,
                    len(paths[Ntrain:]),
                    template=template_eval,
                    augment=None,
                    shuffle_buffer_size=None,
                    num_parallel_calls=num_parallel_calls,
                )
            return ds_train, ds_eval
        raise ValueError(
            "Provided paths did not pass validation. Please "
            "check that they have the same shape, and the "
            "targets have appropriate labels"
        )



import numpy as np
import pytest
import tensorflow as tf

from nobrainer.bayesian_utils import default_mean_field_normal_fn

from ..autoencoder import autoencoder
from ..bayesian_vnet import bayesian_vnet
from ..bayesian_vnet_semi import bayesian_vnet_semi
from ..brainsiam import brainsiam
from ..dcgan import dcgan
from ..highresnet import highresnet
from ..meshnet import meshnet
from ..progressivegan import progressivegan
from ..unet import unet
from ..unet_lstm import unet_lstm
from ..vnet import vnet
from ..vox2vox import Vox_ensembler, vox_gan


def model_test(model_cls, n_classes, input_shape, kwds={}):
    """Tests for models."""
    x = 10 * np.random.random(input_shape)
    y = np.random.choice([True, False], input_shape)

    # Assume every model class has n_classes and input_shape arguments.
    model = model_cls(n_classes=n_classes, input_shape=input_shape[1:], **kwds)
    model.compile(tf.optimizers.Adam(), "binary_crossentropy")
    model.fit(x, y)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape[:-1] + (n_classes,)


def test_highresnet():
    model_test(highresnet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def test_meshnet():
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 37},
    )
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 67},
    )
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 129},
    )
    with pytest.raises(ValueError):
        model_test(
            meshnet,
            n_classes=1,
            input_shape=(1, 32, 32, 32, 1),
            kwds={"receptive_field": 50},
        )


def test_unet():
    model_test(unet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def test_autoencoder():
    """Special test for autoencoder."""

    input_shape = (1, 32, 32, 32, 1)
    x = 10 * np.random.random(input_shape)

    model = autoencoder(input_shape[1:], encoding_dim=128, n_base_filters=32)
    model.compile(tf.optimizers.Adam(), "mse")
    model.fit(x, x)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape


def test_brainsiam():
    """Testing the encoder-projector and predictor structures of the brainsiam architecture"""
    input_shape = (1, 32, 32, 32, 1)
    x = 10 * np.random.random(input_shape)

    n_classes = 1
    weight_decay = 0.0005
    projection_dim = 2048
    latent_dim = 512

    encoder, predictor = brainsiam(
        n_classes,
        input_shape=input_shape[1:],
        weight_decay=weight_decay,
        projection_dim=projection_dim,
        latent_dim=latent_dim,
    )

    encoder_output = encoder(x[1:])
    enc_output_shape = encoder_output.get_shape().as_list()

    predictor_out = predictor(encoder_output)
    pred_output_shape = predictor_out.get_shape().as_list()

    assert (
        enc_output_shape[1] == projection_dim
    ), "encoder output shape not the same as projection dim"
    assert (
        pred_output_shape[1] == projection_dim
    ), "predictor output shape not the same as projection dim"


def test_progressivegan():
    """Test for both discriminator and generator of progressive gan"""

    latent_size = 256
    label_size = 2
    g_fmap_base = 1024
    d_fmap_base = 1024
    alpha = 1.0

    generator, discriminator = progressivegan(
        latent_size,
        label_size=label_size,
        g_fmap_base=g_fmap_base,
        d_fmap_base=d_fmap_base,
    )

    resolutions = [8, 16]

    for res in resolutions:
        generator.add_resolution()
        discriminator.add_resolution()

        latent_input = np.random.random((10, latent_size))
        real_image_input = np.random.random((10, res, res, res, 1))

        fake_images = generator([latent_input, alpha])
        real_pred, real_labels_pred = discriminator([real_image_input, alpha])
        fake_pred, fake_labels_pred = discriminator([fake_images, alpha])

        assert fake_images.shape == real_image_input.shape
        assert real_pred.shape == (real_image_input.shape[0],)
        assert fake_pred.shape == (real_image_input.shape[0],)
        assert real_labels_pred.shape == (real_image_input.shape[0], label_size)
        assert fake_labels_pred.shape == (real_image_input.shape[0], label_size)


def test_dcgan():
    """Special test for dcgan."""

    output_shape = (1, 32, 32, 32, 1)
    z_dim = 32
    z = np.random.random((1, z_dim))

    pred_shape = (1, 8, 8, 8, 1)

    generator, discriminator = dcgan(output_shape[1:], z_dim=z_dim)
    generator.compile(tf.optimizers.Adam(), "mse")
    discriminator.compile(tf.optimizers.Adam(), "mse")

    fake_images = generator.predict(z)
    fake_pred = discriminator.predict(fake_images)

    assert fake_images.shape == output_shape and fake_pred.shape == pred_shape


def test_vnet():
    model_test(vnet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def model_test_bayesian(model_cls, n_classes, input_shape, kernel_posterior_fn):
    """Tests for models."""
    x = 10 * np.random.random(input_shape)
    y = np.random.choice([True, False], input_shape)

    # Assume every model class has n_classes and input_shape arguments.
    model = model_cls(
        n_classes=n_classes,
        input_shape=input_shape[1:],
        kernel_posterior_fn=kernel_posterior_fn,
    )
    model.compile(tf.optimizers.Adam(), "binary_crossentropy")
    model.fit(x, y)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape[:-1] + (n_classes,)


def test_bayesian_vnet_semi():
    model_test_bayesian(
        bayesian_vnet_semi,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kernel_posterior_fn=default_mean_field_normal_fn(weightnorm=True),
    )


def test_bayesian_vnet():
    model_test_bayesian(
        bayesian_vnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kernel_posterior_fn=default_mean_field_normal_fn(weightnorm=True),
    )


def test_unet_lstm():
    input_shape = (1, 32, 32, 32, 32)
    n_classes = 1
    x = 10 * np.random.random(input_shape)
    y = 10 * np.random.random(input_shape)
    model = unet_lstm(input_shape=(32, 32, 32, 32, 1), n_classes=1)
    actual_output = model.predict(x)
    assert actual_output.shape == y.shape[:-1] + (n_classes,)


def test_vox2vox():
    input_shape = (1, 32, 32, 32, 1)
    n_classes = 1
    x = 10 * np.random.random(input_shape)
    y = np.random.choice([True, False], input_shape)

    # testing ensembler
    model_test(Vox_ensembler, n_classes, input_shape)

    # testing Vox2VoxGan
    vox_generator, vox_discriminator = vox_gan(n_classes, input_shape[1:])

    # testing generator
    vox_generator.compile(tf.optimizers.Adam(), "binary_crossentropy")
    vox_generator.fit(x, y)
    actual_output = vox_generator.predict(x)
    assert actual_output.shape == x.shape[:-1] + (n_classes,)

    # testing descriminator
    pred_shape = (1, 2, 2, 2, 1)
    out = vox_discriminator(inputs=[y, x])
    assert out.shape == pred_shape



# import tensorflow as tf
# from tensorflow.python.ops import gen_nn_ops

# def _get_sequence(value, n, channel_index, name):
#   """Formats a value input for gen_nn_ops."""
#   # Performance is fast-pathed for common cases:
#   # `None`, `list`, `tuple` and `int`.
#   if value is None:
#     return [1] * (n + 2)

#   # Always convert `value` to a `list`.
#   if isinstance(value, list):
#     pass
#   elif isinstance(value, tuple):
#     value = list(value)
#   elif isinstance(value, int):
#     value = [value]
#   elif not isinstance(value, collections_abc.Sized):
#     value = [value]
#   else:
#     value = list(value)  # Try casting to a list.

#   len_value = len(value)

#   # Fully specified, including batch and channel dims.
#   if len_value == n + 2:
#     return value

#   # Apply value to spatial dims only.
#   if len_value == 1:
#     value = value * n  # Broadcast to spatial dimensions.
#   elif len_value != n:
#     raise ValueError(f"{name} should be of length 1, {n} or {n + 2}. "
#                      f"Received: {name}={value} of length {len_value}")

#   # Add batch and channel dims (always 1).
#   if channel_index == 1:
#     return [1, 1] + value
#   else:
#     return [1] + value + [1]

# @tf_export("nn.max_pool4d")
# @dispatch.add_dispatch_support
# def max_pool4d(input, ksize, strides, padding, data_format="NVDHWC", name=None):
#     """Performs the max pooling on the input.
#     Args:
#         input: A 6-D `Tensor` of the format specified by `data_format`.
#         ksize: An int or list of `ints` that has length `1`, `3` or `5`. The size of
#         the window for each dimension of the input tensor.
#         strides: An int or list of `ints` that has length `1`, `3` or `5`. The
#           stride of the sliding window for each dimension of the input tensor.
#         padding: A string, either `'VALID'` or `'SAME'`. The padding algorithm. See
#           [here](https://www.tensorflow.org/api_docs/python/tf/nn#notes_on_padding_2)
#           for more information.
#         data_format: An optional string from: "NVDHWC", "NCVDHW". Defaults to "NVDHWC".
#           The data format of the input and output data. With the default format
#           "NVDHWC", the data is stored in the order of: [batch, in_depth, in_height,
#             in_width, in_channels]. Alternatively, the format could be "NCVDHW", the
#           data storage order is: [batch, in_channels, in_volumes, in_depth, in_height,
#             in_width].
#         name: A name for the operation (optional).
#     Returns:
#         A `Tensor` of format specified by `data_format`.
#         The max pooled output tensor.
#   """
#   with ops.name_scope(name, "MaxPool4D", [input]) as name:
#     if data_format is None:
#       data_format = "NVDHWC"
#     channel_index = 1 if data_format.startswith("NC") else 5

#     ksize = _get_sequence(ksize, 3, channel_index, "ksize")
#     strides = _get_sequence(strides, 3, channel_index, "strides")

#     return gen_nn_ops.max_pool4d(
#         input,
#         ksize=ksize,
#         strides=strides,
#         padding=padding,
#         data_format=data_format,
#         name=name)



# Adaptation of the VNet model from https://arxiv.org/pdf/1606.04797.pdf
# This 3D deep neural network model is regularized with 3D spatial dropout
# and Group normalization.

from tensorflow.keras.layers import (
    Conv3D,
    Input,
    MaxPooling3D,
    SpatialDropout3D,
    UpSampling3D,
    concatenate,
)
from tensorflow.keras.models import Model

from ..layers.groupnorm import GroupNormalization


def down_stage(inputs, filters, kernel_size=3, activation="relu", padding="SAME"):
    """encoding block of the VNet model.

    Parameters
    ----------
    inputs: tf.layer for encoding stage.
    filters: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Returns
    ----------
    encoding module.
    """
    convd = Conv3D(filters, kernel_size, activation=activation, padding=padding)(inputs)
    convd = GroupNormalization()(convd)
    convd = Conv3D(filters, kernel_size, activation=activation, padding=padding)(convd)
    convd = GroupNormalization()(convd)
    pool = MaxPooling3D()(convd)
    return convd, pool


def up_stage(inputs, skip, filters, kernel_size=3, activation="relu", padding="SAME"):
    """decoding block of the VNet model.

    Parameters
    ----------
    inputs: tf.layer for encoding stage.
    filters: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Returns
    ----------
    decoded module.
    """
    up = UpSampling3D()(inputs)
    up = Conv3D(filters, 2, activation=activation, padding=padding)(up)
    up = GroupNormalization()(up)

    merge = concatenate([skip, up])
    merge = GroupNormalization()(merge)

    convu = Conv3D(filters, kernel_size, activation=activation, padding=padding)(merge)
    convu = GroupNormalization()(convu)
    convu = Conv3D(filters, kernel_size, activation=activation, padding=padding)(convu)
    convu = GroupNormalization()(convu)
    convu = SpatialDropout3D(0.5)(convu, training=True)

    return convu


def end_stage(inputs, n_classes=1, kernel_size=3, activation="relu", padding="SAME"):
    """last logit layer.

    Parameters
    ----------
    inputs: tf.model layer.
    n_classes: int, for binary class use the value 1.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Result
    ----------
    prediction probablities
    """
    conv = Conv3D(
        filters=n_classes,
        kernel_size=kernel_size,
        activation=activation,
        padding="SAME",
    )(inputs)
    if n_classes == 1:
        conv = Conv3D(n_classes, 1, activation="sigmoid")(conv)
    else:
        conv = Conv3D(n_classes, 1, activation="softmax")(conv)

    return conv


def vnet(
    n_classes=1,
    input_shape=(128, 128, 128, 1),
    kernel_size=3,
    activation="relu",
    padding="SAME",
    **kwargs
):
    """Instantiate a 3D VNet Architecture.

    VNet model: a 3D deep neural network model adapted from
    https://arxiv.org/pdf/1606.04797.pdf adatptations include groupnorm
    and spatial dropout.

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernal of conv layers. Default kernal size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Returns
    ----------
    Model object.

    """
    inputs = Input(input_shape)

    conv1, pool1 = down_stage(
        inputs, 16, kernel_size=kernel_size, activation=activation, padding=padding
    )
    conv2, pool2 = down_stage(
        pool1, 32, kernel_size=kernel_size, activation=activation, padding=padding
    )
    conv3, pool3 = down_stage(
        pool2, 64, kernel_size=kernel_size, activation=activation, padding=padding
    )
    conv4, _ = down_stage(
        pool3, 128, kernel_size=kernel_size, activation=activation, padding=padding
    )
    conv4 = SpatialDropout3D(0.5)(conv4, training=True)

    conv5 = up_stage(
        conv4,
        conv3,
        64,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv6 = up_stage(
        conv5,
        conv2,
        32,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv7 = up_stage(
        conv6,
        conv1,
        16,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )

    conv8 = end_stage(
        conv7,
        n_classes=n_classes,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )

    return Model(inputs=inputs, outputs=conv8)



"""Model definition for DCGAN.
"""
import math

from tensorflow.keras import layers, models


def dcgan(
    output_shape,
    z_dim=256,
    n_base_filters=16,
    batchnorm=True,
    batch_size=None,
    name="dcgan",
):
    """Instantiate DCGAN Architecture.

    Parameters
    ----------
    output_shape: list or tuple of four ints, the shape of the output images. Should be
        scaled to [0,1]. Omit the batch dimension, and include the number of channels.
        Currently, only squares and cubes supported.
    z_dim: int, the dimensions of the encoding of the latent code. This would translate
        to a latent code of dimensions encoding_dimx1.
    n_base_filters: int, number of base filters the models first convolutional layer.
        The subsequent layers have n_filters which are multiples of n_base_filters.
    batchnorm: bool, whether to use batch normalization in the network.
    batch_size: int, number of samples in each batch. This must be set when
        training on TPUs.
    name: str, name to give to the resulting model object.

    Returns
    -------
    Generator Model object.
    Discriminator Model object.
    """

    conv_kwds = {"kernel_size": 4, "activation": None, "padding": "same", "strides": 2}

    conv_transpose_kwds = {
        "kernel_size": 4,
        "strides": 2,
        "activation": None,
        "padding": "same",
    }

    dimensions = output_shape[:-1]
    n_dims = len(dimensions)

    if not (n_dims in [2, 3] and dimensions[1:] == dimensions[:-1]):
        raise ValueError("Dimensions should be of square or cube!")

    Conv = getattr(layers, "Conv{}D".format(n_dims))
    ConvTranspose = getattr(layers, "Conv{}DTranspose".format(n_dims))
    n_layers = int(math.log(dimensions[0], 2))

    # Generator
    z_input = layers.Input(shape=(z_dim,), batch_size=batch_size)

    project = layers.Dense(pow(4, n_dims) * z_dim)(z_input)
    project = layers.ReLU()(project)
    project = layers.Reshape((4,) * n_dims + (z_dim,))(project)
    x = project

    for i in range(n_layers - 2)[::-1]:
        n_filters = min(n_base_filters * (2 ** (i)), z_dim)

        x = ConvTranspose(n_filters, **conv_transpose_kwds)(x)
        if batchnorm:
            x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)

    outputs = Conv(1, 3, activation="sigmoid", padding="same")(x)

    generator = models.Model(
        inputs=[z_input], outputs=[outputs], name=name + "_generator"
    )

    # PatchGAN Discriminator with output of 8x8(x8)
    inputs = layers.Input(shape=(output_shape), batch_size=batch_size)
    x = inputs
    for i in range(n_layers - 3):
        n_filters = min(n_base_filters * (2 ** (i)), z_dim)

        x = Conv(n_filters, **conv_kwds)(x)
        if batchnorm:
            x = layers.BatchNormalization()(x)
        x = layers.ReLU()(x)

    pred = Conv(1, 3, padding="same", activation="sigmoid")(x)

    discriminator = models.Model(
        inputs=[inputs], outputs=[pred], name=name + "_discriminator"
    )

    return generator, discriminator



"""Convolutional variational layers."""

from __future__ import absolute_import, division, print_function

import tensorflow.compat.v2 as tf
from tensorflow.python.layers import utils as tf_layers_util
from tensorflow.python.ops import nn_ops
from tensorflow_probability.python.distributions import independent as independent_lib
from tensorflow_probability.python.distributions import kullback_leibler as kl_lib
from tensorflow_probability.python.distributions import normal as normal_lib
from tensorflow_probability.python.internal import docstring_util
from tensorflow_probability.python.layers import util as tfp_layers_util

__all__ = [
    "Convolution3DReparameterization",
]
doc_args = """activation: Activation function. Set it to None to maintain a
      linear activation.
  activity_regularizer: Regularizer function for the output.
  kernel_posterior_fn: Python `callable` which creates
    `tfd.Distribution` instance representing the surrogate
    posterior of the `kernel` parameter. Default value:
    `default_mean_field_normal_fn()`.
  kernel_posterior_tensor_fn: Python `callable` which takes a
    `tfd.Distribution` instance and returns a representative
    value. Default value: `lambda d: d.sample()`.
  kernel_prior_fn: Python `callable` which creates `tfd`
    instance. See `default_mean_field_normal_fn` docstring for required
    parameter signature.
    Default value: `tfd.Normal(loc=0., scale=1.)`.
  kernel_divergence_fn: Python `callable` which takes the surrogate posterior
    distribution, prior distribution and random variate sample(s) from the
    surrogate posterior and computes or approximates the KL divergence. The
    distributions are `tfd.Distribution`-like instances and the
    sample is a `Tensor`.
  bias_posterior_fn: Python `callable` which creates
    `tfd.Distribution` instance representing the surrogate
    posterior of the `bias` parameter. Default value:
    `default_mean_field_normal_fn(is_singular=True)` (which creates an
    instance of `tfd.Deterministic`).
  bias_posterior_tensor_fn: Python `callable` which takes a
    `tfd.Distribution` instance and returns a representative
    value. Default value: `lambda d: d.sample()`.
  bias_prior_fn: Python `callable` which creates `tfd` instance.
    See `default_mean_field_normal_fn` docstring for required parameter
    signature. Default value: `None` (no prior, no variational inference)
  bias_divergence_fn: Python `callable` which takes the surrogate posterior
    distribution, prior distribution and random variate sample(s) from the
    surrogate posterior and computes or approximates the KL divergence. The
    distributions are `tfd.Distribution`-like instances and the
    sample is a `Tensor`."""


"""Variational convolutional layer Abstract nD convolution layer (private,
  used as implementation base).
  This layer creates a convolution kernel that is convolved
  (actually cross-correlated) with the layer input to produce a tensor of
  outputs. It may also include a bias addition and activation function
  on the outputs. It assumes the `kernel` and/or `bias` are drawn from
  distributions.
  By default, the layer implements a stochastic forward pass via
  sampling from the kernel and bias posteriors,
  ```none
  outputs = f(inputs; kernel, bias), kernel, bias ~ posterior
  ```
  where f denotes the layer's calculation.
  The arguments permit separate specification of the surrogate posterior
  (`q(W|x)`), prior (`p(W)`), and divergence for both the `kernel` and `bias`
  distributions.
"""


class _ConvVariational(tf.keras.layers.Layer):
    @docstring_util.expand_docstring(args=doc_args)
    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        is_mc,
        strides=1,
        padding="valid",
        data_format="channels_last",
        dilation_rate=1,
        activation=None,
        activity_regularizer=None,
        kernel_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(),
        kernel_posterior_tensor_fn=lambda d: d.sample(),
        kernel_prior_fn=tfp_layers_util.default_multivariate_normal_fn,
        kernel_divergence_fn=(lambda q, p, ignore: kl_lib.kl_divergence(q, p)),
        bias_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(
            is_singular=True
        ),
        bias_posterior_tensor_fn=lambda d: d.sample(),
        bias_prior_fn=None,
        bias_divergence_fn=lambda q, p, ignore: kl_lib.kl_divergence(q, p),
        **kwargs
    ):
        super(_ConvVariational, self).__init__(
            activity_regularizer=activity_regularizer, **kwargs
        )
        self.rank = rank
        self.is_mc = is_mc
        self.filters = filters
        self.kernel_size = tf_layers_util.normalize_tuple(
            kernel_size, rank, "kernel_size"
        )
        self.strides = tf_layers_util.normalize_tuple(strides, rank, "strides")
        self.padding = tf_layers_util.normalize_padding(padding)
        self.data_format = tf_layers_util.normalize_data_format(data_format)
        self.dilation_rate = tf_layers_util.normalize_tuple(
            dilation_rate, rank, "dilation_rate"
        )
        self.activation = tf.keras.activations.get(activation)
        self.input_spec = tf.keras.layers.InputSpec(ndim=self.rank + 2)
        self.kernel_posterior_fn = kernel_posterior_fn
        self.kernel_posterior_tensor_fn = kernel_posterior_tensor_fn
        self.kernel_prior_fn = kernel_prior_fn
        self.kernel_divergence_fn = kernel_divergence_fn
        self.bias_posterior_fn = bias_posterior_fn
        self.bias_posterior_tensor_fn = bias_posterior_tensor_fn
        self.bias_prior_fn = bias_prior_fn
        self.bias_divergence_fn = bias_divergence_fn

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        if self.data_format == "channels_first":
            channel_axis = 1
        else:
            channel_axis = -1
        input_dim = tf.compat.dimension_value(input_shape[channel_axis])
        if input_dim is None:
            raise ValueError("The channel dimension of inputs Found `None`.")
        kernel_shape = self.kernel_size + (input_dim, self.filters)
        # If self.dtype is None, build weights using the default dtype.
        dtype = tf.as_dtype(self.dtype or tf.keras.backend.floatx())
        # Must have a posterior kernel.
        self.kernel_posterior = self.kernel_posterior_fn(
            dtype, kernel_shape, "kernel_posterior", self.trainable, self.add_variable
        )
        if self.kernel_prior_fn is None:
            self.kernel_prior = None
        else:
            self.kernel_prior = self.kernel_prior_fn(
                dtype, kernel_shape, "kernel_prior", self.trainable, self.add_variable
            )
        if self.bias_posterior_fn is None:
            self.bias_posterior = None
        else:
            self.bias_posterior = self.bias_posterior_fn(
                dtype,
                (self.filters,),
                "bias_posterior",
                self.trainable,
                self.add_variable,
            )
        if self.bias_prior_fn is None:
            self.bias_prior = None
        else:
            self.bias_prior = self.bias_prior_fn(
                dtype, (self.filters,), "bias_prior", self.trainable, self.add_variable
            )
        self.input_spec = tf.keras.layers.InputSpec(
            ndim=self.rank + 2, axes={channel_axis: input_dim}
        )
        self._convolution_op = nn_ops.Convolution(
            input_shape,
            filter_shape=tf.TensorShape(kernel_shape),
            dilation_rate=self.dilation_rate,
            strides=self.strides,
            padding=self.padding.upper(),
            data_format=tf_layers_util.convert_data_format(
                self.data_format, self.rank + 2
            ),
        )
        self.built = True

    def call(self, inputs):
        inputs = tf.convert_to_tensor(value=inputs, dtype=self.dtype)
        outputs = self._apply_variational_kernel(inputs)
        outputs = self._apply_variational_bias(outputs)
        if self.activation is not None:
            outputs = self.activation(outputs)
        self._apply_divergence(
            self.kernel_divergence_fn,
            self.kernel_posterior,
            self.kernel_prior,
            self.kernel_posterior_tensor,
            name="divergence_kernel",
        )
        self._apply_divergence(
            self.bias_divergence_fn,
            self.bias_posterior,
            self.bias_prior,
            self.bias_posterior_tensor,
            name="divergence_bias",
        )
        return outputs

    def compute_output_shape(self, input_shape):
        """Computes the output shape of the layer.
        Args:
          input_shape: Shape tuple (tuple of integers) or list of shape tuples
            (one per output tensor of the layer). Shape tuples can include None for
            free dimensions, instead of an integer.
        Returns:
          output_shape: A tuple representing the output shape.
        """
        input_shape = tf.TensorShape(input_shape).as_list()
        if self.data_format == "channels_last":
            space = input_shape[1:-1]
            new_space = []
            for i in range(len(space)):
                new_dim = tf_layers_util.conv_output_length(
                    space[i],
                    self.kernel_size[i],
                    padding=self.padding,
                    stride=self.strides[i],
                    dilation=self.dilation_rate[i],
                )
                new_space.append(new_dim)
            return tf.TensorShape([input_shape[0]] + new_space + [self.filters])
        else:
            space = input_shape[2:]
            new_space = []
            for i in range(len(space)):
                new_dim = tf_layers_util.conv_output_length(
                    space[i],
                    self.kernel_size[i],
                    padding=self.padding,
                    stride=self.strides[i],
                    dilation=self.dilation_rate[i],
                )
                new_space.append(new_dim)
            return tf.TensorShape([input_shape[0], self.filters] + new_space)

    def get_config(self):
        """Returns the config of the layer.
        A layer config is a Python dictionary (serializable) containing the
        configuration of a layer. The same layer can be reinstantiated later
        (without its trained weights) from this configuration.
        Returns:
          config: A Python dictionary of class keyword arguments and their
            serialized values.
        """
        config = {
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "activation": (
                tf.keras.activations.serialize(self.activation)
                if self.activation
                else None
            ),
            "activity_regularizer": tf.keras.initializers.serialize(
                self.activity_regularizer
            ),
        }
        function_keys = [
            "kernel_posterior_fn",
            "kernel_posterior_tensor_fn",
            "kernel_prior_fn",
            "kernel_divergence_fn",
            "bias_posterior_fn",
            "bias_posterior_tensor_fn",
            "bias_prior_fn",
            "bias_divergence_fn",
        ]
        for function_key in function_keys:
            function = getattr(self, function_key)
            if function is None:
                function_name = None
                function_type = None
            else:
                function_name, function_type = tfp_layers_util.serialize_function(
                    function
                )
            config[function_key] = function_name
            config[function_key + "_type"] = function_type
        base_config = super(_ConvVariational, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        """Creates a layer from its config.
        This method is the reverse of `get_config`, capable of instantiating the
        same layer from the config dictionary.
        Args:
          config: A Python dictionary, typically the output of `get_config`.
        Returns:
          layer: A layer instance.
        """
        config = config.copy()
        function_keys = [
            "kernel_posterior_fn",
            "kernel_posterior_tensor_fn",
            "kernel_prior_fn",
            "kernel_divergence_fn",
            "bias_posterior_fn",
            "bias_posterior_tensor_fn",
            "bias_prior_fn",
            "bias_divergence_fn",
        ]
        for function_key in function_keys:
            serial = config[function_key]
            function_type = config.pop(function_key + "_type")
            if serial is not None:
                config[function_key] = tfp_layers_util.deserialize_function(
                    serial, function_type=function_type
                )
        return cls(**config)

    def _apply_variational_bias(self, inputs):
        if self.bias_posterior is None:
            self.bias_posterior_tensor = None
            return inputs
        self.bias_posterior_tensor = self.bias_posterior_tensor_fn(self.bias_posterior)
        outputs = inputs
        if self.data_format == "channels_first":
            if self.rank == 1:
                bias = tf.reshape(self.bias_posterior_tensor, [1, self.filters, 1])
                outputs += bias
            if self.rank == 2:
                outputs = tf.nn.bias_add(
                    outputs, self.bias_posterior_tensor, data_format="NCHW"
                )
            if self.rank == 3:
                outputs_shape = outputs.shape.as_list()
                outputs_4d = tf.reshape(
                    outputs,
                    [
                        outputs_shape[0],
                        outputs_shape[1],
                        outputs_shape[2] * outputs_shape[3],
                        outputs_shape[4],
                    ],
                )
                outputs_4d = tf.nn.bias_add(
                    outputs_4d, self.bias_posterior_tensor, data_format="NCHW"
                )
                outputs = tf.reshape(outputs_4d, outputs_shape)
        else:
            outputs = tf.nn.bias_add(
                outputs, self.bias_posterior_tensor, data_format="NHWC"
            )
        return outputs

    def _apply_divergence(
        self, divergence_fn, posterior, prior, posterior_tensor, name
    ):
        if divergence_fn is None or posterior is None or prior is None:
            divergence = None
            return
        divergence = tf.identity(
            divergence_fn(posterior, prior, posterior_tensor), name=name
        )
        self.add_loss(divergence)


class _ConvReparameterization(_ConvVariational):
    """Abstract nD convolution layer (private, used as implementation base).
    This layer creates a convolution kernel that is convolved
    (actually cross-correlated) with the layer input to produce a tensor of
    outputs. It may also include a bias addition and activation function
    on the outputs. It assumes the `kernel` and/or `bias` are drawn from
    distributions.
    By default, the layer implements a stochastic forward pass via
    sampling from the kernel and bias posteriors,
    ```none
    outputs = f(inputs; kernel, bias), kernel, bias ~ posterior
    ```
    where f denotes the layer's calculation. It uses the reparameterization
    estimator [(Kingma and Welling, 2014)][1], which performs a Monte Carlo
    approximation of the distribution integrating over the `kernel` and `bias`.
    The arguments permit separate specification of the surrogate posterior
    (`q(W|x)`), prior (`p(W)`), and divergence for both the `kernel` and `bias`
    distributions.
    #### References
    [1]: Diederik Kingma and Max Welling. Auto-Encoding Variational Bayes. In
         _International Conference on Learning Representations_, 2014.
         https://arxiv.org/abs/1312.6114
    """

    @docstring_util.expand_docstring(args=doc_args)
    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        is_mc,
        strides=1,
        padding="valid",
        data_format="channels_last",
        dilation_rate=1,
        activation=None,
        activity_regularizer=None,
        kernel_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(),
        kernel_posterior_tensor_fn=lambda d: d.sample(),
        kernel_prior_fn=tfp_layers_util.default_multivariate_normal_fn,
        kernel_divergence_fn=lambda q, p, ignore: kl_lib.kl_divergence(q, p),
        bias_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(
            is_singular=True
        ),
        bias_posterior_tensor_fn=lambda d: d.sample(),
        bias_prior_fn=None,
        bias_divergence_fn=lambda q, p, ignore: kl_lib.kl_divergence(q, p),
        **kwargs
    ):
        super(_ConvReparameterization, self).__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            is_mc=is_mc,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=tf.keras.activations.get(activation),
            activity_regularizer=activity_regularizer,
            kernel_posterior_fn=kernel_posterior_fn,
            kernel_posterior_tensor_fn=kernel_posterior_tensor_fn,
            kernel_prior_fn=kernel_prior_fn,
            kernel_divergence_fn=kernel_divergence_fn,
            bias_posterior_fn=bias_posterior_fn,
            bias_posterior_tensor_fn=bias_posterior_tensor_fn,
            bias_prior_fn=bias_prior_fn,
            bias_divergence_fn=bias_divergence_fn,
            **kwargs
        )

    def _apply_variational_kernel(self, inputs):
        if not isinstance(
            self.kernel_posterior, independent_lib.Independent
        ) or not isinstance(self.kernel_posterior.distribution, normal_lib.Normal):
            self.kernel_posterior_tensor = self.kernel_posterior_tensor_fn(
                self.kernel_posterior
            )
            self.kernel_posterior_affine = None
            self.kernel_posterior_affine_tensor = None
            outputs = self._convolution_op(inputs, self.kernel_posterior_tensor)
            return outputs
        else:
            self.kernel_posterior_affine = normal_lib.Normal(
                loc=tf.zeros_like(self.kernel_posterior.distribution.loc),
                scale=self.kernel_posterior.distribution.scale,
            )
            self.kernel_posterior_affine_tensor = self.kernel_posterior_tensor_fn(
                self.kernel_posterior_affine
            )
            self.kernel_posterior_tensor = None
            outputs_m = self._convolution_op(
                inputs, self.kernel_posterior.distribution.loc
            )
            outputs_v = self._convolution_op(
                tf.square(inputs),
                tf.square(self.kernel_posterior.distribution.stddev()),
            )
            k_size = tf_layers_util.normalize_tuple(self.kernel_size, 3, "k_size")
            g_shape = [1 for i in k_size] + [1, self.filters]
            outputs_e = tf.random.normal(shape=g_shape, dtype=self.dtype)
            if self.is_mc:
                err = tf.sqrt(tf.add(outputs_v, tf.keras.backend.epsilon())) * outputs_e
                return outputs_m + err
            else:
                return outputs_m


class Conv3DReparameterization(_ConvReparameterization):
    """3D convolution layer (e.g. spatial convolution over volumes).
    This layer creates a convolution kernel that is convolved
    (actually cross-correlated) with the layer input to produce a tensor of
    outputs. It may also include a bias addition and activation function
    on the outputs. It assumes the `kernel` and/or `bias` are drawn from
    distributions.
    By default, the layer implements a stochastic forward pass via
    sampling from the kernel and bias posteriors,
    ```none
    outputs = f(inputs; kernel, bias), kernel, bias ~ posterior
    ```
    where f denotes the layer's calculation. It uses the reparameterization
    estimator [(Kingma and Welling, 2014)][1], which performs a Monte Carlo
    approximation of the distribution integrating over the `kernel` and `bias`.
    The arguments permit separate specification of the surrogate posterior
    (`q(W|x)`), prior (`p(W)`), and divergence for both the `kernel` and `bias`
    distributions.
    Upon being built, this layer adds losses (accessible via the `losses`
    property) representing the divergences of `kernel` and/or `bias` surrogate
    posteriors and their respective priors. When doing minibatch stochastic
    optimization, make sure to scale this loss such that it is applied just once
    per epoch (e.g. if `kl` is the sum of `losses` for each element of the batch,
    you should pass `kl / num_examples_per_epoch` to your optimizer).
    #### Examples
    We illustrate a Bayesian neural network with [variational inference](
    https://en.wikipedia.org/wiki/Variational_Bayesian_methods),
    assuming a dataset of `features` and `labels`.
    ```python
    import tensorflow as tf
    import tensorflow_probability as tfp
    model = tf.keras.Sequential([
        tf.keras.layers.Reshape([256, 32, 32, 3]),
        tfp.layers.Convolution3DReparameterization(
            64, kernel_size=5, padding='SAME', activation=tf.nn.relu),
        tf.keras.layers.MaxPooling3D(pool_size=[2, 2, 2],
                                     strides=[2, 2, 2],
                                     padding='SAME'),
        tf.keras.layers.Flatten(),
        tfp.layers.DenseReparameterization(10),
    ])
    logits = model(features)
    neg_log_likelihood = tf.nn.softmax_cross_entropy_with_logits(
        labels=labels, logits=logits)
    kl = sum(model.losses)
    loss = neg_log_likelihood + kl
    train_op = tf.train.AdamOptimizer().minimize(loss)
    ```
    It uses reparameterization gradients to minimize the
    Kullback-Leibler divergence up to a constant, also known as the
    negative Evidence Lower Bound. It consists of the sum of two terms:
    the expected negative log-likelihood, which we approximate via
    Monte Carlo; and the KL divergence, which is added via regularizer
    terms which are arguments to the layer.
    #### References
    [1]: Diederik Kingma and Max Welling. Auto-Encoding Variational Bayes. In
         _International Conference on Learning Representations_, 2014.
         https://arxiv.org/abs/1312.6114
    """

    @docstring_util.expand_docstring(args=doc_args)
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1, 1),
        padding="valid",
        data_format="channels_last",
        dilation_rate=(1, 1, 1),
        activation=None,
        activity_regularizer=None,
        is_mc=tf.constant(False, dtype=tf.bool),
        kernel_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(),
        kernel_posterior_tensor_fn=lambda d: d.sample(),
        kernel_prior_fn=tfp_layers_util.default_multivariate_normal_fn,
        kernel_divergence_fn=lambda q, p, ignore: kl_lib.kl_divergence(q, p),
        bias_posterior_fn=tfp_layers_util.default_mean_field_normal_fn(
            is_singular=True
        ),
        bias_posterior_tensor_fn=lambda d: d.sample(),
        bias_prior_fn=None,
        bias_divergence_fn=lambda q, p, ignore: kl_lib.kl_divergence(q, p),
        **kwargs
    ):
        # pylint: disable=g-doc-args
        """Construct layer.
        Args:
          filters: Integer, the dimensionality of the output space (i.e. the number
            of filters in the convolution).
          kernel_size: An integer or tuple/list of 3 integers, specifying the
            depth, height and width of the 3D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
          strides: An integer or tuple/list of 3 integers,
            specifying the strides of the convolution along the depth,
            height and width.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
          padding: One of `"valid"` or `"same"` (case-insensitive).
          data_format: A string, one of `channels_last` (default) or
            `channels_first`. The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape `(batch, depth,
            height, width, channels)` while `channels_first` corresponds to inputs
            with shape `(batch, channels, depth, height, width)`.
          dilation_rate: An integer or tuple/list of 3 integers, specifying
            the dilation rate to use for dilated convolution.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any stride value != 1.
          ${args}"""
        super(Conv3DReparameterization, self).__init__(
            rank=3,
            filters=filters,
            kernel_size=kernel_size,
            is_mc=is_mc,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=tf.keras.activations.get(activation),
            activity_regularizer=activity_regularizer,
            kernel_posterior_fn=kernel_posterior_fn,
            kernel_posterior_tensor_fn=kernel_posterior_tensor_fn,
            kernel_prior_fn=kernel_prior_fn,
            kernel_divergence_fn=kernel_divergence_fn,
            bias_posterior_fn=bias_posterior_fn,
            bias_posterior_tensor_fn=bias_posterior_tensor_fn,
            bias_prior_fn=bias_prior_fn,
            bias_divergence_fn=bias_divergence_fn,
            **kwargs
        )


Convolution3DReparameterization = Conv3DReparameterization



import importlib
import os

import tensorflow as tf

from .base import BaseEstimator
from .. import losses, metrics
from ..dataset import get_steps_per_epoch


class Segmentation(BaseEstimator):
    """Perform segmentation type operations"""

    state_variables = ["block_shape_", "volume_shape_", "scalar_labels_"]

    def __init__(self, base_model, model_args=None):
        if not isinstance(base_model, str):
            self.base_model = base_model.__name__
        else:
            self.base_model = base_model
        self.model_ = None
        self.model_args = model_args or {}
        self.block_shape_ = None
        self.volume_shape_ = None
        self.scalar_labels_ = None

    def fit(
        self,
        dataset_train,
        dataset_validate=None,
        epochs=1,
        checkpoint_dir=os.getcwd(),
        multi_gpu=False,
        warm_start=False,
        # TODO: figure out whether optimizer args should be flattened
        optimizer=None,
        opt_args=None,
        loss=losses.dice,
        metrics=metrics.dice,
    ):
        """Train a segmentation model"""
        # TODO: check validity of datasets

        # extract dataset information
        batch_size = dataset_train.element_spec[0].shape[0]
        self.block_shape_ = tuple(dataset_train.element_spec[0].shape[1:4])
        self.volume_shape_ = dataset_train.volume_shape
        self.scalar_labels_ = True
        n_classes = 1
        if len(dataset_train.element_spec[1].shape) > 1:
            n_classes = dataset_train.element_spec[1].shape[4]
            self.scalar_labels_ = False
        opt_args = opt_args or {}
        if optimizer is None:
            optimizer = tf.keras.optimizers.Adam
            opt_args_tmp = dict(learning_rate=1e-04)
            opt_args_tmp.update(**opt_args)
            opt_args = opt_args_tmp

        def _create(base_model):
            # Instantiate and compile the model
            self.model_ = base_model(
                n_classes=n_classes,
                input_shape=(*self.block_shape_, 1),
                **self.model_args
            )

        def _compile():
            self.model_.compile(
                optimizer(**opt_args),
                loss=loss,
                metrics=metrics,
            )

        if warm_start:
            if self.model is None:
                raise ValueError("warm_start requested, but model is undefined")
            if multi_gpu:
                strategy = tf.distribute.MirroredStrategy()
                with strategy.scope():
                    _compile()
            else:
                _compile()
        else:
            mod = importlib.import_module("..models", "nobrainer.processing")
            base_model = getattr(mod, self.base_model)
            if multi_gpu:
                strategy = tf.distribute.MirroredStrategy()
                with strategy.scope():
                    _create(base_model)
                    _compile()
            else:
                _create(base_model)
                _compile()
        print(self.model_.summary())

        train_steps = get_steps_per_epoch(
            n_volumes=dataset_train.n_volumes,
            volume_shape=self.volume_shape_,
            block_shape=self.block_shape_,
            batch_size=batch_size,
        )

        evaluate_steps = None
        if dataset_validate is not None:
            evaluate_steps = get_steps_per_epoch(
                n_volumes=dataset_validate.n_volumes,
                volume_shape=self.volume_shape_,
                block_shape=self.block_shape_,
                batch_size=batch_size,
            )

        # TODO add checkpoint
        self.model_.fit(
            dataset_train,
            epochs=epochs,
            steps_per_epoch=train_steps,
            validation_data=dataset_validate,
            validation_steps=evaluate_steps,
        )

        return self

    def predict(self, x, batch_size=1, normalizer=None):
        """Makes a prediction using the trained model"""
        if self.model_ is None:
            raise ValueError("Model is undefined. Please train or load a model")
        from ..prediction import predict

        return predict(
            x,
            self.model_,
            block_shape=self.block_shape_,
            batch_size=batch_size,
            normalizer=normalizer,
        )



"""Volume processing methods."""

import itertools

import numpy as np
import tensorflow as tf


def binarize(x):
    """Converts all values greater than 0 to 1 and all others to 0.

    Parameters
    ----------
    x: tensor, values to binarize.

    Returns
    -------
    Tensor of binarized values.
    """
    x = tf.convert_to_tensor(x)
    return tf.cast(x > 0, dtype=x.dtype)


def replace(x, mapping, zero=True):
    """Replace values in tensor `x` using dictionary `mapping`.

    Parameters
    ----------
    x: tensor, values to replace.
    mapping: dict, dictionary mapping original values to new values. Values in
        x equal to a key in the mapping are replaced with the corresponding
        value. Keys and values may overlap.
    zero: boolean, zero values in `x` not in `mapping.keys()`.

    Returns
    -------
    Modified tensor.
    """
    x = tf.cast(x, dtype=tf.int32)
    keys = tf.convert_to_tensor(list(mapping.keys()))
    vals = tf.convert_to_tensor(list(mapping.values()))

    sidx = tf.argsort(keys)
    ks = tf.gather(keys, sidx)
    vs = tf.gather(vals, sidx)

    idx = tf.searchsorted(ks, tf.reshape(x, (-1,)))
    idx = tf.reshape(idx, x.shape)

    # Zero values that are equal to len(vs).
    idx = tf.multiply(idx, tf.cast(tf.not_equal(idx, vs.shape[0]), tf.int32))
    mask = tf.equal(tf.gather(ks, idx), x)
    if zero:
        out = tf.where(mask, tf.gather(vs, idx), 0)
    else:
        out = tf.where(mask, tf.gather(vs, idx), x)

    return out


def standardize(x):
    """Standard score input tensor.

    Implements `(x - mean(x)) / stdev(x)`.

    Parameters
    ----------
    x: tensor, values to standardize.

    Returns
    -------
    Tensor of standardized values. Output has mean 0 and standard deviation 1.
    """
    x = tf.convert_to_tensor(x)
    if x.dtype != tf.float32:
        x = tf.cast(x, tf.float32)
    mean, var = tf.nn.moments(x, axes=None)
    std = tf.sqrt(var)
    return (x - mean) / std


def normalize(x):
    """Normalize the input to 0 and 1.

    Implements `(x - min(x)) / (max(x) - min(x))`.

    Parameters
    ----------
    x: tensor, values to normalize.

    Returns
    -------
    Tensor of normalized values. Output has min 0 and max 1.
    """
    x = tf.convert_to_tensor(x)
    if x.dtype != tf.float32:
        x = tf.cast(x, tf.float32)
    min_x = tf.reduce_min(x)
    max_x = tf.reduce_max(x)
    return (x - min_x) / (max_x - min_x)


def _to_blocks_perm(ndims):
    """Build permutation vector to go from volume to blocks.

    For 3D input, perm will be (0, 2, 4, 1, 3, 5). For 4D input (i.e. 3D volume
    with channels), perm will be (0, 2, 4, 6, 1, 3, 5, 7). Higher dimensional
    input is possible here.

    Parameters
    ----------
    ndims : int
        Number of dimensions in blocks and volumes.

    Returns
    -------
    perm : tuple
        The permutation vector
    """
    perm = np.empty(2 * ndims, dtype=int)
    perm[:ndims] = np.arange(0, ndims * 2, 2)
    perm[ndims:] = np.arange(1, ndims * 2, 2)
    return tuple(perm)


def _from_blocks_perm(ndims):
    """Build permutation vector to go from blocks to volume.

    For 3D input, perm will be (0, 3, 1, 4, 2, 5). For 4D input (i.e. 3D volume
    with channels), perm will be (0, 4, 1, 5, 2, 6, 3, 7). Higher dimensional
    input is possible here.

    Parameters
    ----------
    ndims : int
        Number of dimensions in blocks and volumes.

    Returns
    -------
    perm : tuple
        The permutation vector
    """
    perm = np.empty(2 * ndims, dtype=int)
    perm[::2] = np.arange(ndims)
    perm[1::2] = np.arange(ndims, 2 * ndims)
    return tuple(perm)


def to_blocks(x, block_shape):
    """Split tensor into non-overlapping blocks of shape `block_shape`.

    For the reverse of this function, see `from_blocks`.

    Parameters
    ----------
    x: tensor, 3D tensor to separate into non-overlapping sub-volumes.
    block_shape: tuple of length 3, shape of non-overlapping sub-volumes.

    Returns
    -------
    Tensor with shape `(n_blocks, *block_shape)`.
    """
    x = tf.convert_to_tensor(x)
    volume_shape = np.array(x.shape)

    if isinstance(block_shape, int):
        block_shape = list([block_shape]) * 3
    elif len(block_shape) < 3:
        raise ValueError("expected block_shape to be 1 or >= 3 values.")

    block_shape = np.asarray(block_shape)
    blocks = volume_shape // block_shape

    inter_shape = list(itertools.chain(*zip(blocks, block_shape)))
    new_shape = (-1, *block_shape)
    perm = _to_blocks_perm(ndims=len(block_shape))

    return tf.reshape(
        tf.transpose(tf.reshape(x, shape=inter_shape), perm=perm), shape=new_shape
    )


def from_blocks(x, output_shape):
    """Combine 4D array of non-overlapping sub-volumes `x` into 3D tensor of
    shape `output_shape`.

    For the reverse of this function, see `to_blocks`.

    Parameters
    ----------
    x: tensor, 4D tensor with shape `(N, *block_shape)`, where `N` is the number
        of sub-volumes.
    output_shape: tuple of length 3, shape of resulting volumes.

    Returns
    -------
    Tensor with shape `output_shape`.
    """
    x = tf.convert_to_tensor(x)
    n_blocks = x.shape[0]
    block_shape = x.shape[1:]
    ncbrt = np.cbrt(n_blocks).round(6)
    if not ncbrt.is_integer():
        raise ValueError("Cubed root of number of blocks is not an integer")
    ncbrt = int(ncbrt)
    perm = _from_blocks_perm(ndims=len(block_shape))
    intershape = (ncbrt, ncbrt, ncbrt, *block_shape)
    # Allow channels (i.e. 4D)
    if len(block_shape) == 4:
        intershape = (ncbrt, ncbrt, ncbrt, 1, *block_shape)

    return tf.reshape(
        tf.transpose(tf.reshape(x, shape=intershape), perm=perm), shape=output_shape
    )


def adjust_dynamic_range(x, drange_in, drange_out):
    """Scale and shift tensor.

    Implements `(x * scale) + bias`.

    Parameters
    ----------
    x: array, values to scale and shift.
    drange_in: tuple, input range of values
    drange_out: tuple, output range of values

    Returns
    -------
    Array of scaled and shifted values. Output values has range in drange_out.
    """
    x = tf.convert_to_tensor(x)
    scale = (drange_out[1] - drange_out[0]) / (drange_in[1] - drange_in[0])
    bias = drange_out[0] - drange_in[0] * scale
    return (x * scale) + bias


# Below this line, we implement methods similar to those above but using Numpy.
# This is particularly useful when we use models to predict, because it is
# usually more pleasant to predict on Numpy arrays.


def standardize_numpy(a):
    """Standard score array.

    Implements `(x - mean(x)) / stdev(x)`.

    Parameters
    ----------
    x: array, values to standardize.

    Returns
    -------
    Array of standardized values. Output has mean 0 and standard deviation 1.
    """
    a = np.asarray(a)
    return (a - a.mean()) / a.std()


def normalize_numpy(a):
    """Normalize the array between 0 and 1.

    Implements `(x - min(x)) / (max(x) - min(x))`.

    Parameters
    ----------
    x: array, values to normalize.

    Returns
    -------
    Array of normalized values. Output has min 0 and max 1.
    """
    a = np.asarray(a)
    return (a - a.min()) / (a.max() - a.min())


def from_blocks_numpy(a, output_shape):
    """Combine 4D array of non-overlapping blocks `a` into 3D array of shape
    `output_shape`.

    For the reverse of this function, see `to_blocks_numpy`.

    Parameters
    ----------
    a: array-like, 4D or 5D array of blocks with shape (N, *block_shape), where
        N is the number of blocks.
    output_shape: tuple of len 3 or 4, shape of the combined array.

    Returns
    -------
    Rank 3 array with shape `output_shape`.
    """
    a = np.asarray(a)

    if a.ndim not in [4, 5]:
        raise ValueError("This function only works for 4D or 5D arrays.")
    if len(output_shape) not in [3, 4]:
        raise ValueError("output_shape must have three or four values.")

    n_blocks = a.shape[0]
    block_shape = a.shape[1:]
    ncbrt = np.cbrt(n_blocks).round(6)
    if not ncbrt.is_integer():
        raise ValueError("Cubed root of number of blocks is not an integer")
    ncbrt = int(ncbrt)
    perm = _from_blocks_perm(ndims=len(block_shape))
    intershape = (ncbrt, ncbrt, ncbrt, *block_shape)
    # Allow channels (i.e. 4D)
    if len(block_shape) == 4:
        intershape = (ncbrt, ncbrt, ncbrt, 1, *block_shape)

    return a.reshape(intershape).transpose(perm).reshape(output_shape)


def to_blocks_numpy(a, block_shape):
    """Return new array of non-overlapping blocks of shape `block_shape` from
    array `a`.

    For the reverse of this function (blocks to array), see `from_blocks_numpy`.

    Parameters
    ----------
    a: array-like, 3D or 4D array to block
    block_shape: tuple of len 3 or 4, shape of non-overlapping blocks.

    Returns
    -------
    Rank 4 or 5 array with shape `(N, *block_shape)`, where N is the number of
    blocks.
    """
    a = np.asarray(a)
    orig_shape = np.asarray(a.shape)

    if a.ndim not in [3, 4]:
        raise ValueError("This function only supports 3D or 4D arrays.")

    if isinstance(block_shape, int):
        block_shape = tuple(list([block_shape]) * 3)

    if len(block_shape) not in [3, 4]:
        raise ValueError("block_shape must have three values.")

    blocks = orig_shape // block_shape
    inter_shape = tuple(e for tup in zip(blocks, block_shape) for e in tup)
    new_shape = (-1,) + block_shape
    perm = _to_blocks_perm(ndims=len(block_shape))

    return a.reshape(inter_shape).transpose(perm).reshape(new_shape)


def replace_in_numpy(x, mapping, zero=True):
    """Replace values in numpy ndarray `x` using dictionary `mapping`.

    # Based on https://stackoverflow.com/a/47171600
    """
    # Extract out keys and values
    k = np.array(list(mapping.keys()))
    v = np.array(list(mapping.values()))

    # Get argsort indices
    sidx = k.argsort()

    ks = k[sidx]
    vs = v[sidx]
    idx = np.searchsorted(ks, x)

    idx[idx == len(vs)] = 0
    mask = ks[idx] == x

    if not zero:
        return np.where(mask, vs[idx], x)
    else:
        return np.where(mask, vs[idx], 0)


def adjust_dynamic_range_numpy(a, drange_in, drange_out):
    """Scale and shift numpy array.

    Implements `(a * scale) + bias`.

    Parameters
    ----------
    a: array, values to scale and shift.
    drange_in: tuple, input range of values
    drange_out: tuple, output range of values

    Returns
    -------
    Array of scaled and shifted values. Output values has range in drange_out.
    """
    a = np.asarray(a)
    scale = (drange_out[1] - drange_out[0]) / (drange_in[1] - drange_in[0])
    bias = drange_out[0] - drange_in[0] * scale
    return a * scale + bias



"""Model definition for MeshNet.

Implemented according to the [MeshNet manuscript](https://arxiv.org/abs/1612.00940)
"""

import tensorflow as tf
from tensorflow.keras import layers


def meshnet(
    n_classes,
    input_shape,
    receptive_field=67,
    filters=71,
    activation="relu",
    dropout_rate=0.25,
    batch_size=None,
    name="meshnet",
):
    """Instantiate MeshNet model.

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    receptive_field: {37, 67, 129}, the receptive field of the model. According
        to the MeshNet manuscript, the receptive field should be similar to your
        input shape. The actual receptive field is the cube of the value provided.
    filters: int, number of filters per volumetric convolution. The original
        MeshNet manuscript uses 21 filters for a binary segmentation task
        (i.e., brain extraction) and 71 filters for a multi-class segmentation task.
    activation: str or optimizer object, the non-linearity to use.
    dropout_rate: float between 0 and 1, the fraction of input units to drop.
    batch_size: int, number of samples in each batch. This must be set when
        training on TPUs.
    name: str, name to give to the resulting model object.

    Returns
    -------
    Model object.

    Raises
    ------
    ValueError if receptive field is not an allowable value.
    """

    if receptive_field not in {37, 67, 129}:
        raise ValueError("unknown receptive field. Legal values are 37, 67, and 129.")

    def one_layer(x, layer_num, dilation_rate=(1, 1, 1)):
        x = layers.Conv3D(
            filters,
            kernel_size=(3, 3, 3),
            padding="same",
            dilation_rate=dilation_rate,
            name="layer{}/conv3d".format(layer_num),
        )(x)
        x = layers.BatchNormalization(name="layer{}/batchnorm".format(layer_num))(x)
        x = layers.Activation(activation, name="layer{}/activation".format(layer_num))(
            x
        )
        x = layers.Dropout(dropout_rate, name="layer{}/dropout".format(layer_num))(x)
        return x

    inputs = layers.Input(shape=input_shape, batch_size=batch_size, name="inputs")

    if receptive_field == 37:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3)
        x = one_layer(x, 4, dilation_rate=(2, 2, 2))
        x = one_layer(x, 5, dilation_rate=(4, 4, 4))
        x = one_layer(x, 6, dilation_rate=(8, 8, 8))
        x = one_layer(x, 7)
    elif receptive_field == 67:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3, dilation_rate=(2, 2, 2))
        x = one_layer(x, 4, dilation_rate=(4, 4, 4))
        x = one_layer(x, 5, dilation_rate=(8, 8, 8))
        x = one_layer(x, 6, dilation_rate=(16, 16, 16))
        x = one_layer(x, 7)
    elif receptive_field == 129:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2, dilation_rate=(2, 2, 2))
        x = one_layer(x, 3, dilation_rate=(4, 4, 4))
        x = one_layer(x, 4, dilation_rate=(8, 8, 8))
        x = one_layer(x, 5, dilation_rate=(16, 16, 16))
        x = one_layer(x, 6, dilation_rate=(32, 32, 32))
        x = one_layer(x, 7)

    x = layers.Conv3D(
        filters=n_classes,
        kernel_size=(1, 1, 1),
        padding="same",
        name="classification/conv3d",
    )(x)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    x = layers.Activation(final_activation, name="classification/activation")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



import numpy as np
import pytest

from nobrainer import spatial_transforms as transformations


@pytest.fixture(scope="session")
def test_centercrop():
    # Test for inputs
    shape = (10, 10, 10)
    x = np.ones(shape).astype(np.float32)
    y = np.random.randint(0, 2, size=shape).astype(np.float32)
    fine = int(x.shape[1])
    x = transformations.centercrop(x, finesize=fine)
    x = x.numpy()
    # Test for output shapes
    assert x.shape[1] == fine & x.shape[0] == fine & x.shape[2] == shape[2]
    assert y.shape[1] == fine & y.shape[0] == fine & y.shape[2] == shape[2]

    # test for both x and y
    shape = (10, 10, 10)
    x = np.ones(shape).astype(np.float32)
    y = np.random.randint(0, 2, size=shape).astype(np.float32)
    fine = int(x.shape[1])
    x, y = transformations.centercrop(x, y, fine, trans_xy=True)
    x = x.numpy()
    y = y.numpy()
    # Test for output shapes
    assert x.shape[1] == fine & x.shape[0] == fine & x.shape[2] == shape[2]
    assert y.shape[1] == fine & y.shape[0] == fine & y.shape[2] == shape[2]

    # Test for varying finesize
    shape = (10, 10, 10)
    x = np.ones(shape).astype(np.float32)
    y = np.random.randint(0, 2, size=shape).astype(np.float32)
    finesize = [128, 1]
    x1, y1 = transformations.centercrop(x, y, finesize[0], trans_xy=True)
    x2, y2 = transformations.centercrop(x, y, finesize[1], trans_xy=True)
    x1 = x1.numpy()
    x2 = x2.numpy()
    y1 = y1.numpy()
    y2 = y2.numpy()
    assert (
        x1.shape[1]
        == min(shape[1], finesize[0]) & x1.shape[0]
        == min(shape[0], finesize[0]) & x1.shape[2]
        == shape[2]
    )
    assert (
        y1.shape[1]
        == min(shape[1], finesize[0]) & y1.shape[0]
        == min(shape[0], finesize[0]) & y1.shape[2]
        == shape[2]
    )

    assert (
        x2.shape[1]
        == min(shape[1], finesize[1]) & x2.shape[0]
        == min(shape[0], finesize[1])
    )
    assert (
        y2.shape[1]
        == min(shape[1], finesize[1]) & y2.shape[0]
        == min(shape[0], finesize[1])
    )
    assert y2.shape[2] == shape[2] & x2.shape[2] == shape[2]


def test_spatialConstantPadding():
    x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]], [[4, 4, 4], [5, 5, 5], [6, 6, 6]]])
    y = np.array([[[1, 0, 1], [0, 2, 2], [3, 3, 0]], [[4, 1, 4], [5, 0, 0], [0, 0, 0]]])
    x_expected = np.array(
        [
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 3.0, 3.0, 3.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                [0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0],
                [0.0, 0.0, 6.0, 6.0, 6.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
        ]
    )
    y_expected = np.array(
        [
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 2.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 4.0, 1.0, 4.0, 0.0, 0.0],
                [0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
        ]
    )
    resultx, resulty = transformations.spatialConstantPadding(
        x, y, trans_xy=True, padding_zyx=[0, 2, 2]
    )
    np.testing.assert_allclose(x_expected, resultx.numpy())
    np.testing.assert_allclose(y_expected, resulty.numpy())


def test_randomCrop():
    x = np.random.rand(10, 10, 10).astype(np.float32)
    y = np.random.randint(0, 2, size=(10, 10, 10)).astype(np.float32)
    expected_shape = (3, 3, 10)
    res_x, res_y = transformations.randomCrop(x, y, trans_xy=True, cropsize=3)
    assert np.shape(res_x.numpy()) == expected_shape
    assert np.shape(res_y.numpy()) == expected_shape
    assert np.all(np.in1d(np.ravel(res_x), np.ravel(x)))
    assert np.all(np.in1d(np.ravel(res_y), np.ravel(y)))


def test_resize():
    x = np.random.rand(10, 10, 10).astype(np.float32)
    y = np.random.randint(0, 2, size=(10, 10, 10)).astype(np.float32)
    expected_shape = (5, 5, 10)
    results_x, results_y = transformations.resize(
        x, y, trans_xy=True, size=[5, 5], mode="bicubic"
    )
    assert np.shape(results_x.numpy()) == expected_shape
    assert np.shape(results_y.numpy()) == expected_shape


def test_randomflip_leftright():
    x = np.random.rand(3, 3, 3).astype(np.float32)
    y = np.random.randint(0, 2, size=(3, 3, 3)).astype(np.float32)
    res_x, res_y = transformations.randomflip_leftright(x, y, trans_xy=True)
    expected_shape = (3, 3, 3)
    assert np.shape(res_x.numpy()) == expected_shape
    assert np.shape(res_y.numpy()) == expected_shape
